@@ -1,16 +1,21 @@
-#![cfg(feature = "NSString")]
 #![cfg(feature = "NSProcessInfo")]
 use alloc::format;
 
-use crate::Foundation::NSProcessInfo;
+use crate::NSProcessInfo;
 
 #[test]
 fn debug() {
     let info = NSProcessInfo::processInfo();
+
+    #[cfg(feature = "NSString")]
     let expected = format!(
         "NSProcessInfo {{ processName: {:?}, .. }}",
         info.processName()
     );
+
+    #[cfg(not(feature = "NSString"))]
+    let expected = "NSProcessInfo { .. }";
+
     assert_eq!(format!("{info:?}"), expected);
 }
 
